@@ -1,16 +1,14 @@
+from __future__ import print_function
+
 import unittest
 import dbus
 import subprocess
 import os
 import time
-import re
-import shutil
 import sys
-import multiprocessing
 from datetime import datetime
-from enum import Enum
 from systemd import journal
-from time import monotonic
+from monotonic import monotonic
 
 import gi
 gi.require_version('GUdev', '1.0')
@@ -19,9 +17,6 @@ from gi.repository import GUdev
 test_devs = None
 FLIGHT_RECORD_FILE = "flight_record.log"
 
-multiprocessing.set_start_method("fork")
-
-
 def run_command(command):
     res = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
@@ -33,49 +28,26 @@ def run_command(command):
         output = out.decode().strip()
     return (res.returncode, output)
 
-
 def get_call_long(call):
     def call_long(*args, **kwargs):
         """Do an async call with a very long timeout (unless specified otherwise)"""
-        if 'timeout' not in kwargs:
-            kwargs['timeout'] = 100  # seconds
+        kwargs['timeout'] = 100  # seconds
         return call(*args, **kwargs)
 
     return call_long
 
 
-def get_version_from_lsb():
-    ret, out = run_command("lsb_release -rs")
-    if ret != 0:
-        raise RuntimeError("Cannot get distro version from lsb_release output: '%s'" % out)
+def skip_on(skip_on_distros, skip_on_version="", reason=""):
+    """A function returning a decorator to skip some test on a given distribution-version combination
 
-    return out.split(".")[0]
+    :param skip_on_distros: distro(s) to skip the test on
+    :type skip_on_distros: str or tuple of str
+    :param str skip_on_version: version of distro(s) to skip the tests on (only
+                                checked on distribution match)
 
-
-def get_version_from_pretty_name(pretty_name):
-    """ Try to get distro and version from 'OperatingSystemPrettyName'
-        hostname property.
-
-        It should look like this:
-         - "Debian GNU/Linux 9 (stretch)"
-         - "Fedora 27 (Workstation Edition)"
-         - "CentOS Linux 7 (Core)"
-
-        So just return first word as distro and first number as version.
-    """
-    distro = pretty_name.split()[0].lower()
-    match = re.search(r"\d+", pretty_name)
-    if match is not None:
-        version = match.group(0)
-    else:
-        version = get_version_from_lsb()
-
-    return (distro, version)
-
-
-def get_version():
-    """ Try to get distro and version
     """
+    if isinstance(skip_on_distros, str):
+        skip_on_distros = (skip_on_distros,)
 
     bus = dbus.SystemBus()
 
@@ -83,16 +55,39 @@ def get_version():
     sys_info = bus.get_object("org.freedesktop.hostname1", "/org/freedesktop/hostname1")
     cpe = str(sys_info.Get("org.freedesktop.hostname1", "OperatingSystemCPEName", dbus_interface=dbus.PROPERTIES_IFACE))
 
-    if cpe:
-        # 2nd to 4th fields from e.g. "cpe:/o:fedoraproject:fedora:25" or "cpe:/o:redhat:enterprise_linux:7.3:GA:server"
-        _project, distro, version = tuple(cpe.split(":")[2:5])
-        # we want just the major version, so remove all decimal places (if any)
-        version = str(int(float(version)))
-    else:
-        pretty_name = str(sys_info.Get("org.freedesktop.hostname1", "OperatingSystemPrettyName", dbus_interface=dbus.PROPERTIES_IFACE))
-        distro, version = get_version_from_pretty_name(pretty_name)
+    # 2nd to 4th fields from e.g. "cpe:/o:fedoraproject:fedora:25" or "cpe:/o:redhat:enterprise_linux:7.3:GA:server"
+    project, distro, version = tuple(cpe.split(":")[2:5])
 
-    return (distro, version)
+    def decorator(func):
+        if distro in skip_on_distros and (not skip_on_version or skip_on_version == version):
+            msg = "not supported on this distribution in this version" + (": %s" % reason if reason else "")
+            return unittest.skip(msg)(func)
+        else:
+            return func
+
+    return decorator
+
+def unstable_test(test):
+    """Decorator for unstable tests
+
+    Failures of tests decorated with this decorator are silently ignored unless
+    the ``UNSTABLE_TESTS_FATAL`` environment variable is defined.
+    """
+
+    def decorated_test(*args):
+        try:
+            test(*args)
+        except unittest.SkipTest:
+            # make sure skipped tests are just skipped as usual
+            raise
+        except:
+            # and swallow everything else, just report a failure of an unstable
+            # test, unless told otherwise
+            if "UNSTABLE_TESTS_FATAL" in os.environ:
+                raise
+            print("unstable-fail...", end="", file=sys.stderr)
+
+    return decorated_test
 
 
 class DBusProperty(object):
@@ -115,7 +110,7 @@ class DBusProperty(object):
     def _update_value(self):
         self._value = self.obj.Get(self.iface, self.prop, dbus_interface=dbus.PROPERTIES_IFACE)
 
-    def _check(self, timeout, check_fn, poll_vg=None):
+    def _check(self, timeout, check_fn):
         for _ in range(int(timeout / 0.5)):
             try:
                 self._update_value()
@@ -125,37 +120,22 @@ class DBusProperty(object):
                 # ignore all exceptions -- they might be result of property
                 # not having the expected type (e.g. 'None' when checking for len)
                 pass
-            if poll_vg:
-                poll_vg.Poll(dbus_interface='org.freedesktop.UDisks2.VolumeGroup')
             time.sleep(0.5)
 
         return False
 
-    def assertEqual(self, value, timeout=TIMEOUT, getter=None, poll_vg=None, msg=None):
+    def assertEqual(self, value, timeout=TIMEOUT, getter=None):
         if getter is not None:
             check_fn = lambda x: getter(x) == value
         else:
             check_fn = lambda x: x == value
-        ret = self._check(timeout, check_fn, poll_vg)
-
-        if not ret:
-            if getter is not None:
-                raise AssertionError('%s != %s%s' % (getter(self._value), value, ' (%s)' % msg if msg else ''))
-            else:
-                raise AssertionError('%s != %s%s' % (self._value, value, ' (%s)' % msg if msg else ''))
-
-    def assertNotEqual(self, value, timeout=TIMEOUT, getter=None, msg=None):
-        if getter is not None:
-            check_fn = lambda x: getter(x) != value
-        else:
-            check_fn = lambda x: x != value
         ret = self._check(timeout, check_fn)
 
         if not ret:
             if getter is not None:
-                raise AssertionError('%s != %s%s' % (getter(self._value), value, ' (%s)' % msg if msg else ''))
+                raise AssertionError('%s != %s' % (getter(self._value), value))
             else:
-                raise AssertionError('%s != %s%s' % (self._value, value, ' (%s)' % msg if msg else ''))
+                raise AssertionError('%s != %s' % (self._value, value))
 
     def assertAlmostEqual(self, value, delta, timeout=TIMEOUT, getter=None):
         if getter is not None:
@@ -172,47 +152,48 @@ class DBusProperty(object):
                 raise AssertionError('%s is not almost equal to %s (delta = %s)' % (self._value,
                                                                                     value, delta))
 
-    def assertGreater(self, value, timeout=TIMEOUT, msg=None):
+
+    def assertGreater(self, value, timeout=TIMEOUT):
         check_fn = lambda x: x > value
         ret = self._check(timeout, check_fn)
 
         if not ret:
-            raise AssertionError('%s is not greater than %s%s' % (self._value, value, ' (%s)' % msg if msg else ''))
+            raise AssertionError('%s is not grater than %s' % (self._value, value))
 
-    def assertLess(self, value, timeout=TIMEOUT, msg=None):
+    def assertLess(self, value, timeout=TIMEOUT):
         check_fn = lambda x: x < value
         ret = self._check(timeout, check_fn)
 
         if not ret:
-            raise AssertionError('%s is not less than %s%s' % (self._value, value, ' (%s)' % msg if msg else ''))
+            raise AssertionError('%s is not less than %s' % (self._value, value))
 
-    def assertIn(self, lst, timeout=TIMEOUT, msg=None):
+    def assertIn(self, lst, timeout=TIMEOUT):
         check_fn = lambda x: x in lst
         ret = self._check(timeout, check_fn)
 
         if not ret:
             raise AssertionError('%s not found in %s' % (self._value, lst))
 
-    def assertNotIn(self, lst, timeout=TIMEOUT, msg=None):
+    def assertNotIn(self, lst, timeout=TIMEOUT):
         check_fn = lambda x: x not in lst
         ret = self._check(timeout, check_fn)
 
         if not ret:
-            raise AssertionError('%s unexpectedly found in %s%s' % (self._value, lst, ' (%s)' % msg if msg else ''))
+            raise AssertionError('%s unexpectedly found in %s' % (self._value, lst))
 
-    def assertTrue(self, timeout=TIMEOUT, msg=None):
+    def assertTrue(self, timeout=TIMEOUT):
         check_fn = lambda x: bool(x)
         ret = self._check(timeout, check_fn)
 
         if not ret:
-            raise AssertionError('%s is not true%s' % (self._value, ' (%s)' % msg if msg else ''))
+            raise AssertionError('%s is not true' % self._value)
 
-    def assertFalse(self, timeout=TIMEOUT, msg=None):
+    def assertFalse(self, timeout=TIMEOUT):
         check_fn = lambda x: not bool(x)
         ret = self._check(timeout, check_fn)
 
         if not ret:
-            raise AssertionError('%s is not false%s' % (self._value, ' (%s)' % msg if msg else ''))
+            raise AssertionError('%s is not false' % self._value)
 
     def assertIsNone(self, timeout=TIMEOUT):
         check_fn = lambda x: x is None
@@ -221,14 +202,14 @@ class DBusProperty(object):
         if not ret:
             raise AssertionError('%s is not None' % self._value)
 
-    def assertIsNotNone(self, timeout=TIMEOUT, msg=None):
+    def assertIsNotNone(self, timeout=TIMEOUT):
         check_fn = lambda x: x is not None
         ret = self._check(timeout, check_fn)
 
         if not ret:
-            raise AssertionError('unexpectedly None%s' % ' (%s)' % msg if msg else '')
+            raise AssertionError('unexpectedly None')
 
-    def assertLen(self, length, timeout=TIMEOUT, msg=None):
+    def assertLen(self, length, timeout=TIMEOUT):
         check_fn = lambda x: len(x) == length
         ret = self._check(timeout, check_fn)
 
@@ -236,16 +217,9 @@ class DBusProperty(object):
             if not hasattr(self._value, '__len__'):
                 raise AssertionError('%s has no length' % type(self._value))
             else:
-                raise AssertionError('Expected length %d, but %s has length %d%s' % (length,
-                                                                                     self._value,
-                                                                                     len(self._value,
-                                                                                     ' (%s)' % msg if msg else '')))
-    def assertContains(self, member, timeout=TIMEOUT, msg=None):
-        check_fn = lambda x: member in x
-        ret = self._check(timeout, check_fn)
-
-        if not ret:
-            raise AssertionError('%s does not contain %s%s' % (self._value, member, ' (%s)' % msg if msg else ''))
+                raise AssertionError('Expected length %d, but %s has length %d' % (length,
+                                                                                   self._value,
+                                                                                   len(self._value)))
 
 
 class UdisksTestCase(unittest.TestCase):
@@ -256,26 +230,34 @@ class UdisksTestCase(unittest.TestCase):
     distro = (None, None, None)       # (project, distro_name, version)
     no_options = dbus.Dictionary(signature="sv")
 
+
     @classmethod
     def setUpClass(self):
         self.iface_prefix = 'org.freedesktop.UDisks2'
         self.path_prefix = '/org/freedesktop/UDisks2'
         self.bus = dbus.SystemBus()
 
-        self.distro = get_version()
+        # get information about the distribution from systemd (hostname1)
+        sys_info = self.bus.get_object("org.freedesktop.hostname1", "/org/freedesktop/hostname1")
+        cpe = str(sys_info.Get("org.freedesktop.hostname1", "OperatingSystemCPEName", dbus_interface=dbus.PROPERTIES_IFACE))
+
+        # 2nd to 4th fields from e.g. "cpe:/o:fedoraproject:fedora:25" or "cpe:/o:redhat:enterprise_linux:7.3:GA:server"
+        self.distro = tuple(cpe.split(":")[2:5])
 
         self._orig_call_async = self.bus.call_async
         self._orig_call_blocking = self.bus.call_blocking
         self.bus.call_async = get_call_long(self._orig_call_async)
         self.bus.call_blocking = get_call_long(self._orig_call_blocking)
         self.vdevs = test_devs
-        assert len(self.vdevs) > 3
+        assert len(self.vdevs) > 3;
+
 
     @classmethod
     def tearDownClass(self):
         self.bus.call_async = self._orig_call_async
         self.bus.call_blocking = self._orig_call_blocking
 
+
     def run(self, *args):
         record = []
         now = datetime.now()
@@ -323,10 +305,12 @@ class UdisksTestCase(unittest.TestCase):
             obj = self.get_object(obj)
         return dbus.Interface(obj, self.iface_prefix + iface_suffix)
 
+
     @classmethod
     def get_property(self, obj, iface_suffix, prop):
         return DBusProperty(obj, self.iface_prefix + iface_suffix, prop)
 
+
     @classmethod
     def get_property_raw(self, obj, iface_suffix, prop):
         res = obj.Get(self.iface_prefix + iface_suffix, prop, dbus_interface=dbus.PROPERTIES_IFACE)
@@ -358,50 +342,19 @@ class UdisksTestCase(unittest.TestCase):
 
         return False
 
-    def try_unmount(self, path):
-        """Handle unmount and retry if busy"""
-        for _ in range(10):
-            ret, out = self.run_command('umount %s' % path)
-            # the mount may either be unmounted already or not exist anymore
-            if ret == 0 or "not mounted" in out or "no mount point specified" in out or "mountpoint not found" in out:
-                return
-            if "target is busy" not in out:
-                break
-            time.sleep(0.5)
-        self.fail('Failed to unmount %s: %s' % (path, out))
-
-    def _conf_backup(self, conf_file):
-        """ Backup and restore @conf_file during cleanup. If @conf_file doesn't exist, it
-            will be removed during cleanup
-        """
-        if os.path.exists(conf_file):
-            # conf file exists -> backup and restore
-            contents = self.read_file(conf_file)
-            self.addCleanup(self.write_file, conf_file, contents)
-        else:
-            # conf file doesn't exist -> remove during cleanup
-            self.addCleanup(shutil.rmtree, conf_file, True)
-
     @classmethod
     def read_file(self, filename):
         with open(filename, 'r') as f:
             content = f.read()
         return content
 
-    @classmethod
-    def write_file(self, filename, content, ignore_nonexistent=False, binary=False):
-        try:
-            with open(filename, 'wb' if binary else 'w') as f:
-                f.write(content)
-        except OSError as e:
-            if not ignore_nonexistent:
-                raise e
 
     @classmethod
-    def remove_file(self, filename, ignore_nonexistent=False):
+    def write_file(self, filename, content, ignore_nonexistent=False):
         try:
-            os.remove(filename)
-        except OSError as e:
+            with open(filename, 'w') as f:
+                f.write(content)
+        except FileNotFoundError as e:
             if not ignore_nonexistent:
                 raise e
 
@@ -409,28 +362,21 @@ class UdisksTestCase(unittest.TestCase):
     def run_command(self, command):
         return run_command(command)
 
-    @classmethod
-    def module_available(cls, module):
-        ret, _out = cls.run_command('modprobe %s' % module)
-        return ret == 0
-
     @classmethod
     def check_module_loaded(self, module):
-        """Tries to load specified module. No checks for extra Manager interface are done.
-           Returns False when module is not available, True when the module initialized
-           successfully, raises an exception otherwise.
-        """
         manager_obj = self.get_object('/Manager')
         manager = self.get_interface(manager_obj, '.Manager')
-        try:
-            manager.EnableModule(module, dbus.Boolean(True))
+        manager_intro = dbus.Interface(manager_obj, "org.freedesktop.DBus.Introspectable")
+        intro_data = manager_intro.Introspect()
+        modules_loaded = 'interface name="org.freedesktop.UDisks2.Manager.%s"' % module in intro_data
+
+        if not modules_loaded:
+            manager.EnableModules(dbus.Boolean(True))
+            intro_data = manager_intro.Introspect()
+            return 'interface name="org.freedesktop.UDisks2.Manager.%s"' % module in intro_data
+        else:
             return True
-        except dbus.exceptions.DBusException as e:
-            msg = r"Error initializing module '%s': .*\.so: cannot open shared object file: No such file or directory" % module
-            if re.search(msg, e.get_dbus_message()):
-                return False
-            else:
-                raise
+
 
     @classmethod
     def ay_to_str(self, ay):
@@ -439,32 +385,16 @@ class UdisksTestCase(unittest.TestCase):
         return ''.join(chr(x) for x in ay[:-1])
 
     @classmethod
-    def str_to_ay(self, string, terminate=True):
+    def str_to_ay(self, string):
         """Convert a string to a bytearray (terminated with '\0')"""
 
-        if terminate:
-            string += '\0'
+        string += '\0'
 
         return dbus.Array([dbus.Byte(ord(c)) for c in string],
                           signature=dbus.Signature('y'), variant_level=1)
 
     @classmethod
-    def bytes_to_ay(self, bytes):
-        """Convert Python bytes to a DBus bytearray"""
-
-        return dbus.Array([dbus.Byte(b) for b in bytes],
-                          signature=dbus.Signature('y'), variant_level=1)
-
-    @classmethod
-    def set_udev_properties(self, device, props):
-        """Sets one or more udev properties for the 'device' identified by its serial number.
-           Note that this overwrites previously set properties. Pass props=None to remove
-           the rules.
-
-        :type props: dict
-        """
-        UDISKS_UDEV_RULES = "/run/udev/rules.d/99-udisks_test.rules"
-
+    def set_udev_property(self, device, prop, value):
         udev = GUdev.Client()
         dev = udev.query_by_device_file(device)
         serial = dev.get_property("ID_SERIAL")
@@ -475,36 +405,27 @@ class UdisksTestCase(unittest.TestCase):
             # already exists
             pass
 
-        if props:
-            rules = ""
-            for i in props:
-                rules += ', ENV{%s}="%s"' % (i, props[i])
-            self.write_file(UDISKS_UDEV_RULES,
-                            'ENV{ID_SERIAL}=="%s"%s\n' % (serial, rules))
-        else:
-            self.remove_file(UDISKS_UDEV_RULES, ignore_nonexistent=True)
+        self.write_file("/run/udev/rules.d/99-udisks_test.rules",
+                        'ENV{ID_SERIAL}=="%s", ENV{%s}="%s"\n' % (serial, prop, value))
         self.run_command("udevadm control --reload")
         uevent_path = os.path.join(dev.get_sysfs_path(), "uevent")
         self.write_file(uevent_path, "change\n")
         self.udev_settle()
-        # FIXME: need to give udisksd some time to process the uevent
-        time.sleep(1)
+        os.unlink("/run/udev/rules.d/99-udisks_test.rules")
+        self.run_command("udevadm control --reload")
 
     @classmethod
     def assertHasIface(self, obj, iface):
+        obj_intro = dbus.Interface(obj, "org.freedesktop.DBus.Introspectable")
+        intro_data = obj_intro.Introspect()
+
         for _ in range(20):
-            obj_intro = dbus.Interface(obj, "org.freedesktop.DBus.Introspectable")
-            intro_data = obj_intro.Introspect()
-            if 'interface name="%s"' % iface in intro_data:
+            if ('interface name="%s"' % iface) in intro_data:
                 return
             time.sleep(0.5)
 
         raise AssertionError("Object '%s' has no interface '%s'" % (obj.object_path, iface))
 
-    def assertStartswith(self, val, prefix):
-        if not val.startswith(prefix):
-            raise AssertionError("'%s' does not start with '%s'" % (val, prefix))
-
 
 class FlightRecorder(object):
     """Context manager for recording data/logs
@@ -543,7 +464,6 @@ class FlightRecorder(object):
         # given as arguments was not handled
         return False
 
-
 class CmdFlightRecorder(FlightRecorder):
     """Flight recorder running a command and gathering its standard and error output"""
 
@@ -572,7 +492,6 @@ class CmdFlightRecorder(FlightRecorder):
         rec = '<<<<< ' + self._desc + ' >>>>>' + '\n' + out.decode() + '\n\n'
         self._store.append(rec)
 
-
 class JournalRecorder(FlightRecorder):
     """Flight recorder for gathering logs (journal)"""
 
@@ -588,19 +507,20 @@ class JournalRecorder(FlightRecorder):
         self._stopped = None
 
     def _start(self):
-        self._started = time.time()
+        self._started = monotonic()
 
     def _stop(self):
-        self._stopped = time.time()
+        self._stopped = monotonic()
 
     def _save(self):
-        j = journal.Reader(converters={"MESSAGE": lambda x: x.decode(errors="replace")})
+        j = journal.Reader()
         j.this_boot()
-        j.seek_realtime(self._started)
+        j.seek_monotonic(self._started)
         journal_data = ""
 
         entry = j.get_next()
-        while entry and int(entry["__REALTIME_TIMESTAMP"].timestamp()) <= int(self._stopped):
+        # entry["__MONOTONIC_TIMESTAMP"] is a tuple of (datetime.timedelta, boot_id)
+        while entry and entry["__MONOTONIC_TIMESTAMP"][0].seconds <= int(self._stopped):
             if "_COMM" in entry and "_PID" in entry:
                 source = "%s[%d]" % (entry["_COMM"], entry["_PID"])
             else:
@@ -611,38 +531,3 @@ class JournalRecorder(FlightRecorder):
             entry = j.get_next()
         rec = '<<<<< ' + self._desc + ' >>>>>' + '\n' + journal_data + '\n\n\n'
         self._store.append(rec)
-
-
-class TestTags(Enum):
-    ALL = "all"               # "default" tag for running all tests
-    SLOW = "slow"             # slow tests
-    UNSTABLE = "unstable"     # randomly failing tests
-    UNSAFE = "unsafe"         # tests that change system configuration
-    NOSTORAGE = "nostorage"   # tests that don't work with storage
-    EXTRADEPS = "extradeps"   # tests that require special configuration and/or device to run
-
-    @classmethod
-    def get_tags(cls):
-        return [t.value for t in cls.__members__.values()]
-
-    @classmethod
-    def get_tag_by_value(cls, value):
-        tag = next((t for t in cls.__members__.values() if t.value == value), None)
-
-        if not tag:
-            raise ValueError('Unknown value "%s"' % value)
-
-        return tag
-
-
-def tag_test(*tags):
-    def decorator(func):
-        func.slow = TestTags.SLOW in tags
-        func.unstable = TestTags.UNSTABLE in tags
-        func.unsafe = TestTags.UNSAFE in tags
-        func.nostorage = TestTags.NOSTORAGE in tags
-        func.extradeps = TestTags.EXTRADEPS in tags
-
-        return func
-
-    return decorator
